@@ -38,11 +38,27 @@ enum { MIN_GIF_DELAY = 25 };
 float zoom_min;
 float zoom_max;
 
-int zoomdiff(float z1, float z2)
+static int zoomdiff(float z1, float z2)
 {
 	return (int) (z1 * 1000.0 - z2 * 1000.0);
 }
 
+static void img_apply_gamma(img_t *img)
+{
+	if (img == NULL || img->im == NULL || img->cmod == NULL)
+		return;
+	
+	if (img->gamma == 0) {
+		imlib_context_set_color_modifier(NULL);
+	} else {
+		double range = img->gamma <= 0 ? 1.0 : GAMMA_MAX - 1.0;
+
+		imlib_context_set_color_modifier(img->cmod);
+		imlib_reset_color_modifier();
+		imlib_modify_color_modifier_gamma(1.0 + img->gamma * (range / GAMMA_RANGE));
+	}
+}
+
 void img_init(img_t *img, win_t *win)
 {
 	zoom_min = zoom_levels[0] / 100.0;
@@ -57,15 +73,23 @@ void img_init(img_t *img, win_t *win)
 
 	img->im = NULL;
 	img->win = win;
+	img->scalemode = options->scalemode;
 	img->zoom = options->zoom;
 	img->zoom = MAX(img->zoom, zoom_min);
 	img->zoom = MIN(img->zoom, zoom_max);
 	img->checkpan = false;
 	img->dirty = false;
-	img->aa = options->aa;
-	img->alpha = true;
+	img->aa = ANTI_ALIAS;
+	img->alpha = ALPHA_LAYER;
 	img->multi.cap = img->multi.cnt = 0;
 	img->multi.animate = false;
+	img->multi.length = img->multi.repeat = 0;
+
+	img->cmod = imlib_create_color_modifier();
+	img->gamma = MIN(MAX(options->gamma, -GAMMA_RANGE), GAMMA_RANGE);
+	
+	img->ss.on = options->slideshow > 0;
+	img->ss.delay = options->slideshow > 0 ? options->slideshow : SLIDESHOW_DELAY;
 }
 
 void exif_auto_orientate(const fileinfo_t *file)
@@ -76,21 +100,17 @@ void exif_auto_orientate(const fileinfo_t *file)
 		case 2:
 			imlib_image_flip_vertical();
 			break;
-
 		case 3:
 			imlib_image_orientate(2);
 			break;
-
 		case 7:
 			imlib_image_orientate(1);
 		case 4:
 			imlib_image_flip_horizontal();
 			break;
-
 		case 6:
 			imlib_image_orientate(1);
 			break;
-
 		case 8:
 			imlib_image_orientate(3);
 			break;
@@ -106,7 +126,7 @@ bool img_load_gif(img_t *img, const fileinfo_t *file)
 	ColorMapObject *cmap;
 	DATA32 bgpixel, *data, *ptr;
 	DATA32 *prev_frame = NULL;
-	Imlib_Image *im;
+	Imlib_Image im;
 	int i, j, bg, r, g, b;
 	int x, y, w, h, sw, sh;
 	int px, py, pw, ph;
@@ -122,8 +142,8 @@ bool img_load_gif(img_t *img, const fileinfo_t *file)
 		img->multi.frames = (img_frame_t*)
 		                    s_malloc(sizeof(img_frame_t) * img->multi.cap);
 	}
-	img->multi.cnt = 0;
-	img->multi.sel = 0;
+	img->multi.cnt = img->multi.sel = 0;
+	img->multi.length = img->multi.repeat = 0;
 
 #if defined(GIFLIB_MAJOR) && GIFLIB_MAJOR >= 5
 	gif = DGifOpenFileName(file->path, NULL);
@@ -150,7 +170,7 @@ bool img_load_gif(img_t *img, const fileinfo_t *file)
 
 			DGifGetExtension(gif, &ext_code, &ext);
 			while (ext) {
-				if (ext_code == 0xf9) {
+				if (ext_code == GRAPHICS_EXT_FUNC_CODE) {
 					if (ext[1] & 1)
 						transp = (int) ext[4];
 					else
@@ -161,6 +181,12 @@ bool img_load_gif(img_t *img, const fileinfo_t *file)
 						delay = MAX(delay, MIN_GIF_DELAY);
 
 					disposal = (unsigned int) ext[1] >> 2 & 0x7;
+				} else if (ext_code == APPLICATION_EXT_FUNC_CODE) {
+					if (ext[0] == 11 && memcmp(ext+1, "NETSCAPE2.0", 11) == 0) {
+						DGifGetExtensionNext(gif, &ext);
+						if (ext && ext[0] == 3 && ext[1] == 1)
+							img->multi.repeat = ((int) ext[3] << 8 | (int) ext[2]) - 1;
+					}
 				}
 				ext = NULL;
 				DGifGetExtensionNext(gif, &ext);
@@ -247,6 +273,7 @@ bool img_load_gif(img_t *img, const fileinfo_t *file)
 			}
 			img->multi.frames[img->multi.cnt].im = im;
 			img->multi.frames[img->multi.cnt].delay = delay ? delay : GIF_DELAY;
+			img->multi.length += img->multi.frames[img->multi.cnt].delay;
 			img->multi.cnt++;
 		}
 	} while (rec != TERMINATE_RECORD_TYPE);
@@ -302,11 +329,11 @@ bool img_load(img_t *img, const fileinfo_t *file)
 		img_load_gif(img, file);
 #endif
 
+	img_apply_gamma(img);
+
 	img->w = imlib_image_get_width();
 	img->h = imlib_image_get_height();
-	img->scalemode = options->scalemode;
-	img->re = false;
-	img->checkpan = false;
+	img->checkpan = true;
 	img->dirty = true;
 
 	return true;
@@ -334,36 +361,38 @@ void img_close(img_t *img, bool decache)
 			imlib_free_image();
 		img->im = NULL;
 	}
+
+	if (img->cmod)
+		imlib_context_set_color_modifier(NULL);
 }
 
 void img_check_pan(img_t *img, bool moved)
 {
 	win_t *win;
 	int ox, oy;
+	float w, h;
 
 	if (img == NULL || img->im == NULL || img->win == NULL)
 		return;
 
 	win = img->win;
+	w = img->w * img->zoom;
+	h = img->h * img->zoom;
 	ox = img->x;
 	oy = img->y;
 
-	if (img->w * img->zoom > win->w) {
-		if (img->x > 0 && img->x + img->w * img->zoom > win->w)
-			img->x = 0;
-		if (img->x < 0 && img->x + img->w * img->zoom < win->w)
-			img->x = win->w - img->w * img->zoom;
-	} else {
-		img->x = (win->w - img->w * img->zoom) / 2;
-	}
-	if (img->h * img->zoom > win->h) {
-		if (img->y > 0 && img->y + img->h * img->zoom > win->h)
-			img->y = 0;
-		if (img->y < 0 && img->y + img->h * img->zoom < win->h)
-			img->y = win->h - img->h * img->zoom;
-	} else {
-		img->y = (win->h - img->h * img->zoom) / 2;
-	}
+	if (w < win->w)
+		img->x = (win->w - w) / 2;
+	else if (img->x > 0)
+		img->x = 0;
+	else if (img->x + w < win->w)
+		img->x = win->w - w;
+	if (h < win->h)
+		img->y = (win->h - h) / 2;
+	else if (img->y > 0)
+		img->y = 0;
+	else if (img->y + h < win->h)
+		img->y = win->h - h;
 
 	if (!moved && (ox != img->x || oy != img->y))
 		img->dirty = true;
@@ -411,6 +440,8 @@ void img_render(img_t *img)
 	win_t *win;
 	int sx, sy, sw, sh;
 	int dx, dy, dw, dh;
+	Imlib_Image bg;
+	unsigned long c;
 
 	if (img == NULL || img->im == NULL || img->win == NULL)
 		return;
@@ -418,19 +449,6 @@ void img_render(img_t *img)
 	win = img->win;
 	img_fit(img);
 
-	if (!img->re) {
-		/* rendered for the first time */
-		img->re = true;
-		if (img->zoom * img->w <= win->w)
-			img->x = (win->w - img->w * img->zoom) / 2;
-		else
-			img->x = 0;
-		if (img->zoom * img->h <= win->h)
-			img->y = (win->h - img->h * img->zoom) / 2;
-		else
-			img->y = 0;
-	}
-	
 	if (img->checkpan) {
 		img_check_pan(img, false);
 		img->checkpan = false;
@@ -439,8 +457,11 @@ void img_render(img_t *img)
 	if (!img->dirty)
 		return;
 
-	/* calculate source and destination offsets */
-	if (img->x < 0) {
+	/* calculate source and destination offsets:
+	 *   - part of image drawn on full window, or
+	 *   - full image drawn on part of window
+	 */
+	if (img->x <= 0) {
 		sx = -img->x / img->zoom;
 		sw = win->w / img->zoom;
 		dx = 0;
@@ -451,7 +472,7 @@ void img_render(img_t *img)
 		dx = img->x;
 		dw = img->w * img->zoom;
 	}
-	if (img->y < 0) {
+	if (img->y <= 0) {
 		sy = -img->y / img->zoom;
 		sh = win->h / img->zoom;
 		dy = 0;
@@ -467,40 +488,61 @@ void img_render(img_t *img)
 
 	imlib_context_set_image(img->im);
 	imlib_context_set_anti_alias(img->aa);
-
-	if (!img->alpha && imlib_image_has_alpha())
-		win_draw_rect(win, win->pm, dx, dy, dw, dh, True, 0, win->white);
-	
 	imlib_context_set_drawable(win->pm);
-	imlib_render_image_part_on_drawable_at_size(sx, sy, sw, sh, dx, dy, dw, dh);
 
+	if (imlib_image_has_alpha()) {
+		if ((bg = imlib_create_image(dw, dh)) == NULL)
+			die("could not allocate memory");
+		imlib_context_set_image(bg);
+		imlib_image_set_has_alpha(0);
+
+		if (img->alpha) {
+			int i, c, r;
+			DATA32 col[2] = { 0xFF666666, 0xFF999999 };
+			DATA32 * data = imlib_image_get_data();
+
+			for (r = 0; r < dh; r++) {
+				i = r * dw;
+				if (r == 0 || r == 8) {
+					for (c = 0; c < dw; c++)
+						data[i++] = col[!(c & 8) ^ !r];
+				} else {
+					memcpy(&data[i], &data[(r & 8) * dw], dw * sizeof(data[0]));
+				}
+			}
+			imlib_image_put_back_data(data);
+		} else {
+			c = win->fullscreen ? win->fscol : win->bgcol;
+			imlib_context_set_color(c >> 16 & 0xFF, c >> 8 & 0xFF, c & 0xFF, 0xFF);
+			imlib_image_fill_rectangle(0, 0, dw, dh);
+		}
+		imlib_blend_image_onto_image(img->im, 0, sx, sy, sw, sh, 0, 0, dw, dh);
+		imlib_context_set_color_modifier(NULL);
+		imlib_render_image_on_drawable(dx, dy);
+
+		imlib_free_image();
+		if (img->gamma != 0)
+			imlib_context_set_color_modifier(img->cmod);
+	} else {
+		imlib_render_image_part_on_drawable_at_size(sx, sy, sw, sh, dx, dy, dw, dh);
+	}
 	img->dirty = false;
 }
 
 bool img_fit_win(img_t *img, scalemode_t sm)
 {
-	if (img == NULL || img->im == NULL)
-		return false;
-
-	img->scalemode = sm;
-	return img_fit(img);
-}
-
-bool img_center(img_t *img)
-{
-	int ox, oy;
+	float oz;
 
 	if (img == NULL || img->im == NULL || img->win == NULL)
 		return false;
-	
-	ox = img->x;
-	oy = img->y;
 
-	img->x = (img->win->w - img->w * img->zoom) / 2;
-	img->y = (img->win->h - img->h * img->zoom) / 2;
-	
-	if (ox != img->x || oy != img->y) {
-		img->dirty = true;
+	oz = img->zoom;
+	img->scalemode = sm;
+
+	if (img_fit(img)) {
+		img->x = img->win->w / 2 - (img->win->w / 2 - img->x) * img->zoom / oz;
+		img->y = img->win->h / 2 - (img->win->h / 2 - img->y) * img->zoom / oz;
+		img->checkpan = true;
 		return true;
 	} else {
 		return false;
@@ -650,56 +692,59 @@ bool img_pan_edge(img_t *img, direction_t dir)
 	}
 }
 
-void img_rotate(img_t *img, int d)
+void img_rotate(img_t *img, degree_t d)
 {
-	win_t *win;
-	int ox, oy, tmp;
+	int i, ox, oy, tmp;
 
 	if (img == NULL || img->im == NULL || img->win == NULL)
 		return;
 
-	win = img->win;
-	ox = d == 1 ? img->x : win->w - img->x - img->w * img->zoom;
-	oy = d == 3 ? img->y : win->h - img->y - img->h * img->zoom;
-
 	imlib_context_set_image(img->im);
 	imlib_image_orientate(d);
 
-	img->x = oy + (win->w - win->h) / 2;
-	img->y = ox + (win->h - win->w) / 2;
+	for (i = 0; i < img->multi.cnt; i++) {
+		if (i != img->multi.sel) {
+			imlib_context_set_image(img->multi.frames[i].im);
+			imlib_image_orientate(d);
+		}
+	}
+	if (d == DEGREE_90 || d == DEGREE_270) {
+		ox = d == DEGREE_90  ? img->x : img->win->w - img->x - img->w * img->zoom;
+		oy = d == DEGREE_270 ? img->y : img->win->h - img->y - img->h * img->zoom;
 
-	tmp = img->w;
-	img->w = img->h;
-	img->h = tmp;
+		img->x = oy + (img->win->w - img->win->h) / 2;
+		img->y = ox + (img->win->h - img->win->w) / 2;
 
-	img->checkpan = true;
+		tmp = img->w;
+		img->w = img->h;
+		img->h = tmp;
+		img->checkpan = true;
+	}
 	img->dirty = true;
 }
 
-void img_rotate_left(img_t *img)
-{
-	img_rotate(img, 3);
-}
-
-void img_rotate_right(img_t *img)
-{
-	img_rotate(img, 1);
-}
-
 void img_flip(img_t *img, flipdir_t d)
 {
-	if (img == NULL || img->im == NULL)
+	int i;
+	void (*imlib_flip_op[3])(void) = {
+		imlib_image_flip_horizontal,
+		imlib_image_flip_vertical,
+		imlib_image_flip_diagonal
+	};
+
+	d = (d & (FLIP_HORIZONTAL | FLIP_VERTICAL)) - 1;
+
+	if (img == NULL || img->im == NULL || d < 0 || d >= ARRLEN(imlib_flip_op))
 		return;
 
 	imlib_context_set_image(img->im);
+	imlib_flip_op[d]();
 
-	switch (d) {
-		case FLIP_HORIZONTAL:
-			imlib_image_flip_horizontal();
-			break;
-		case FLIP_VERTICAL:
-			imlib_image_flip_vertical();
-			break;
+	for (i = 0; i < img->multi.cnt; i++) {
+		if (i != img->multi.sel) {
+			imlib_context_set_image(img->multi.frames[i].im);
+			imlib_flip_op[d]();
+		}
 	}
 	img->dirty = true;
 }
@@ -715,6 +760,34 @@ void img_toggle_antialias(img_t *img)
 	img->dirty = true;
 }
 
+bool img_change_gamma(img_t *img, int d)
+{
+	/* d < 0: decrease gamma
+	 * d = 0: reset gamma
+	 * d > 0: increase gamma
+	 */
+	int gamma;
+
+	if (img == NULL || img->im == NULL)
+		return false;
+
+	if (d == 0)
+		gamma = 0;
+	else if (d < 0)
+		gamma = MAX(-GAMMA_RANGE, img->gamma - 1);
+	else
+		gamma = MIN(+GAMMA_RANGE, img->gamma + 1);
+
+	if (img->gamma != gamma) {
+		img->gamma = gamma;
+		img_apply_gamma(img);
+		img->dirty = true;
+		return true;
+	} else {
+		return false;
+	}
+}
+
 bool img_frame_goto(img_t *img, int n)
 {
 	if (img == NULL || img->im == NULL)
@@ -754,7 +827,11 @@ bool img_frame_animate(img_t *img, bool restart)
 		return false;
 
 	if (img->multi.sel + 1 >= img->multi.cnt) {
-		if (restart || GIF_LOOP) {
+		if (restart || GIF_LOOP == 1) {
+			img_frame_goto(img, 0);
+		} else if (GIF_LOOP == -1 && img->multi.repeat != 0) {
+			if (img->multi.repeat > 0)
+				img->multi.repeat--;
 			img_frame_goto(img, 0);
 		} else {
 			img->multi.animate = false;
