@@ -31,6 +31,7 @@
 #include <sys/time.h>
 #include <sys/wait.h>
 #include <X11/keysym.h>
+#include <X11/XF86keysym.h>
 
 #include "types.h"
 #include "commands.h"
@@ -46,6 +47,11 @@ enum {
 	TITLE_LEN    = 256
 };
 
+typedef struct {
+	const char *name;
+	char *cmd;
+} exec_t;
+
 typedef struct {
 	struct timeval when;
 	bool active;
@@ -56,6 +62,7 @@ typedef struct {
 void redraw(void);
 void reset_cursor(void);
 void animate(void);
+void slideshow(void);
 void clear_resize(void);
 
 appmode_t mode;
@@ -68,21 +75,27 @@ int filecnt, fileidx;
 int alternate;
 
 int prefix;
+bool extprefix;
 
 bool resized = false;
 
-const char * const INFO_SCRIPT = ".sxiv/exec/image-info";
 struct {
-  char *script;
+  char *cmd;
   int fd;
   unsigned int i, lastsep;
   bool open;
 } info;
 
+struct {
+	char *cmd;
+	bool warned;
+} keyhandler;
+
 timeout_t timeouts[] = {
-	{ { 0, 0 }, false, redraw },
+	{ { 0, 0 }, false, redraw       },
 	{ { 0, 0 }, false, reset_cursor },
-	{ { 0, 0 }, false, animate },
+	{ { 0, 0 }, false, animate      },
+	{ { 0, 0 }, false, slideshow    },
 	{ { 0, 0 }, false, clear_resize },
 };
 
@@ -114,16 +127,28 @@ void check_add_file(char *filename)
 		filecnt *= 2;
 		files = (fileinfo_t*) s_realloc(files, filecnt * sizeof(fileinfo_t));
 	}
+
+#if defined _BSD_SOURCE || defined _XOPEN_SOURCE && \
+    ((_XOPEN_SOURCE - 0) >= 500 || defined _XOPEN_SOURCE_EXTENDED)
+
+	if ((files[fileidx].path = realpath(filename, NULL)) == NULL) {
+		warn("could not get real path of file: %s\n", filename);
+		return;
+	}
+#else
 	if (*filename != '/') {
-		files[fileidx].path = absolute_path(filename);
-		if (files[fileidx].path == NULL) {
+		if ((files[fileidx].path = absolute_path(filename)) == NULL) {
 			warn("could not get absolute path of file: %s\n", filename);
 			return;
 		}
+	} else {
+		files[fileidx].path = NULL;
 	}
+#endif
+
 	files[fileidx].loaded = false;
 	files[fileidx].name = s_strdup(filename);
-	if (*filename == '/')
+	if (files[fileidx].path == NULL)
 		files[fileidx].path = files[fileidx].name;
 	if ((bn = strrchr(files[fileidx].name , '/')) != NULL && bn[1] != '\0')
 		files[fileidx].base = ++bn;
@@ -159,6 +184,8 @@ void remove_file(int n, bool manual)
 	filecnt--;
 	if (n < tns.cnt)
 		tns.cnt--;
+	if (n < alternate)
+		alternate--;
 }
 
 void set_timeout(timeout_f handler, int time, bool overwrite)
@@ -219,7 +246,7 @@ void open_info(void)
 	static pid_t pid;
 	int pfd[2];
 
-	if (info.script == NULL || info.open || win.bar.h == 0)
+	if (info.cmd == NULL || info.open || win.bar.h == 0)
 		return;
 	if (info.fd != -1) {
 		close(info.fd);
@@ -240,8 +267,8 @@ void open_info(void)
 	} else if (pid == 0) {
 		close(pfd[0]);
 		dup2(pfd[1], 1);
-		execl(info.script, info.script, files[fileidx].name, NULL);
-		warn("could not exec: %s", info.script);
+		execl(info.cmd, info.cmd, files[fileidx].name, NULL);
+		warn("could not exec: %s", info.cmd);
 		exit(EXIT_FAILURE);
 	}
 }
@@ -275,6 +302,7 @@ end:
 	info.i -= info.lastsep;
 	win.bar.l[info.i] = '\0';
 	win_update_bar(&win);
+	close(info.fd);
 	info.fd = -1;
 	while (waitpid(-1, NULL, WNOHANG) > 0);
 }
@@ -285,16 +313,20 @@ void load_image(int new)
 		return;
 
 	win_set_cursor(&win, CURSOR_WATCH);
+	reset_timeout(slideshow);
+
+	if (new != fileidx)
+		alternate = fileidx;
 
 	img_close(&img, false);
 	while (!img_load(&img, &files[new])) {
 		remove_file(new, false);
 		if (new >= filecnt)
 			new = filecnt - 1;
+		else if (new > 0 && new < fileidx)
+			new--;
 	}
-
 	files[new].loaded = true;
-	alternate = fileidx;
 	fileidx = new;
 
 	info.open = false;
@@ -312,6 +344,7 @@ void update_info(void)
 	unsigned int i, fn, fw, n;
 	unsigned int llen = sizeof(win.bar.l), rlen = sizeof(win.bar.r);
 	char *lt = win.bar.l, *rt = win.bar.r, title[TITLE_LEN];
+	const char * mark;
 	bool ow_info;
 
 	for (fw = 0, i = filecnt; i > 0; fw++, i /= 10);
@@ -328,9 +361,10 @@ void update_info(void)
 	/* update bar contents */
 	if (win.bar.h == 0)
 		return;
+	mark = files[sel].marked ? "* " : "";
 	if (mode == MODE_THUMB) {
 		if (tns.cnt == filecnt) {
-			n = snprintf(rt, rlen, "%0*d/%d", fw, sel + 1, filecnt);
+			n = snprintf(rt, rlen, "%s%0*d/%d", mark, fw, sel + 1, filecnt);
 			ow_info = true;
 		} else {
 			snprintf(lt, llen, "Loading... %0*d/%d", fw, tns.cnt, filecnt);
@@ -338,14 +372,19 @@ void update_info(void)
 			ow_info = false;
 		}
 	} else {
-		n = snprintf(rt, rlen, "%3d%% | ", (int) (img.zoom * 100.0));
+		n = snprintf(rt, rlen, "%s", mark);
+		if (img.ss.on)
+			n += snprintf(rt + n, rlen - n, "%ds | ", img.ss.delay);
+		if (img.gamma != 0)
+			n += snprintf(rt + n, rlen - n, "G%+d | ", img.gamma);
+		n += snprintf(rt + n, rlen - n, "%3d%% | ", (int) (img.zoom * 100.0));
 		if (img.multi.cnt > 0) {
 			for (fn = 0, i = img.multi.cnt; i > 0; fn++, i /= 10);
 			n += snprintf(rt + n, rlen - n, "%0*d/%d | ",
 			              fn, img.multi.sel + 1, img.multi.cnt);
 		}
 		n += snprintf(rt + n, rlen - n, "%0*d/%d", fw, sel + 1, filecnt);
-		ow_info = info.script == NULL;
+		ow_info = info.cmd == NULL;
 	}
 	if (ow_info) {
 		fn = strlen(files[sel].name);
@@ -362,10 +401,19 @@ void update_info(void)
 
 void redraw(void)
 {
-	if (mode == MODE_IMAGE)
+	int t;
+
+	if (mode == MODE_IMAGE) {
 		img_render(&img);
-	else
+		if (img.ss.on) {
+			t = img.ss.delay * 1000;
+			if (img.multi.cnt > 0 && img.multi.animate)
+				t = MAX(t, img.multi.length);
+			set_timeout(slideshow, t, false);
+		}
+	} else {
 		tns_render(&tns);
+	}
 	update_info();
 	win_draw(&win);
 	reset_timeout(redraw);
@@ -402,54 +450,131 @@ void animate(void)
 	}
 }
 
+void slideshow(void)
+{
+	load_image(fileidx + 1 < filecnt ? fileidx + 1 : 0);
+	redraw();
+}
+
 void clear_resize(void)
 {
 	resized = false;
 }
 
-bool keymask(const keymap_t *k, unsigned int state)
+void run_key_handler(const char *key, unsigned int mask)
 {
-	return (k->ctrl ? ControlMask : 0) == (state & ControlMask);
+	pid_t pid;
+	int retval, status, n = mode == MODE_IMAGE ? fileidx : tns.sel;
+	char kstr[32];
+	struct stat oldst, newst;
+
+	if (keyhandler.cmd == NULL) {
+		if (!keyhandler.warned) {
+			warn("key handler not installed");
+			keyhandler.warned = true;
+		}
+		return;
+	}
+	if (key == NULL)
+		return;
+
+	snprintf(kstr, sizeof(kstr), "%s%s%s%s",
+	         mask & ControlMask ? "C-" : "",
+	         mask & Mod1Mask    ? "M-" : "",
+	         mask & ShiftMask   ? "S-" : "", key);
+
+	stat(files[n].path, &oldst);
+
+	if ((pid = fork()) == 0) {
+		execl(keyhandler.cmd, keyhandler.cmd, kstr, files[n].path, NULL);
+		warn("could not exec key handler");
+		exit(EXIT_FAILURE);
+	} else if (pid < 0) {
+		warn("could not fork key handler");
+		return;
+	}
+	win_set_cursor(&win, CURSOR_WATCH);
+
+	waitpid(pid, &status, 0);
+	retval = WEXITSTATUS(status);
+	if (WIFEXITED(status) == 0 || retval != 0)
+		warn("key handler exited with non-zero return value: %d", retval);
+
+	if (stat(files[n].path, &newst) == 0 &&
+	    memcmp(&oldst.st_mtime, &newst.st_mtime, sizeof(oldst.st_mtime)) == 0)
+	{
+		/* file has not changed */
+		win_set_cursor(&win, CURSOR_ARROW);
+		set_timeout(reset_cursor, TO_CURSOR_HIDE, true);
+		return;
+	}
+	if (mode == MODE_IMAGE) {
+		img_close(&img, true);
+		load_image(fileidx);
+	}
+	if (!tns_load(&tns, n, &files[n], true, mode == MODE_IMAGE) &&
+	    mode == MODE_THUMB)
+	{
+		remove_file(tns.sel, false);
+		tns.dirty = true;
+		if (tns.sel >= tns.cnt)
+			tns.sel = tns.cnt - 1;
+	}
+	redraw();
 }
 
-bool buttonmask(const button_t *b, unsigned int state)
-{
-	return ((b->ctrl ? ControlMask : 0) | (b->shift ? ShiftMask : 0)) ==
-	       (state & (ControlMask | ShiftMask));
-}
+#define MODMASK(mask) ((mask) & (ShiftMask|ControlMask|Mod1Mask))
 
 void on_keypress(XKeyEvent *kev)
 {
 	int i;
-	KeySym ksym;
+	unsigned int sh;
+	KeySym ksym, shksym;
 	char key;
 
 	if (kev == NULL)
 		return;
 
-	XLookupString(kev, &key, 1, &ksym, NULL);
-
-	if ((ksym == XK_Escape || (key >= '0' && key <= '9')) &&
-	    (kev->state & ControlMask) == 0)
-	{
-		/* number prefix for commands */
-		prefix = ksym == XK_Escape ? 0 : prefix * 10 + (int) (key - '0');
-		return;
+	if (kev->state & ShiftMask) {
+		kev->state &= ~ShiftMask;
+		XLookupString(kev, &key, 1, &shksym, NULL);
+		kev->state |= ShiftMask;
 	}
+	XLookupString(kev, &key, 1, &ksym, NULL);
+	sh = (kev->state & ShiftMask) && ksym != shksym ? ShiftMask : 0;
 
-	for (i = 0; i < ARRLEN(keys); i++) {
-		if (keys[i].ksym == ksym && keymask(&keys[i], kev->state)) {
-			if (keys[i].cmd != NULL && keys[i].cmd(keys[i].arg))
+	if (IsModifierKey(ksym))
+		return;
+	if (ksym == XK_Escape && MODMASK(kev->state) == 0) {
+		extprefix = False;
+	} else if (extprefix) {
+		run_key_handler(XKeysymToString(ksym), kev->state & ~sh);
+		extprefix = False;
+	} else if (key >= '0' && key <= '9') {
+		/* number prefix for commands */
+		prefix = prefix * 10 + (int) (key - '0');
+		return;
+	} else for (i = 0; i < ARRLEN(keys); i++) {
+		if (keys[i].ksym == ksym &&
+		    MODMASK(keys[i].mask | sh) == MODMASK(kev->state) &&
+		    keys[i].cmd != NULL)
+		{
+			cmdreturn_t ret = keys[i].cmd(keys[i].arg);
+
+			if (ret == CMD_INVALID)
+				continue;
+			if (ret == CMD_DIRTY)
 				redraw();
-			prefix = 0;
-			return;
+			break;
 		}
 	}
+	prefix = 0;
 }
 
 void on_buttonpress(XButtonEvent *bev)
 {
 	int i, sel;
+	static Time firstclick;
 
 	if (bev == NULL)
 		return;
@@ -460,11 +585,16 @@ void on_buttonpress(XButtonEvent *bev)
 
 		for (i = 0; i < ARRLEN(buttons); i++) {
 			if (buttons[i].button == bev->button &&
-			    buttonmask(&buttons[i], bev->state))
+			    MODMASK(buttons[i].mask) == MODMASK(bev->state) &&
+			    buttons[i].cmd != NULL)
 			{
-				if (buttons[i].cmd != NULL && buttons[i].cmd(buttons[i].arg))
+				cmdreturn_t ret = buttons[i].cmd(buttons[i].arg);
+
+				if (ret == CMD_INVALID)
+					continue;
+				if (ret == CMD_DIRTY)
 					redraw();
-				return;
+				break;
 			}
 		}
 	} else {
@@ -472,17 +602,27 @@ void on_buttonpress(XButtonEvent *bev)
 		switch (bev->button) {
 			case Button1:
 				if ((sel = tns_translate(&tns, bev->x, bev->y)) >= 0) {
-					if (sel == tns.sel) {
-						mode = MODE_IMAGE;
-						set_timeout(reset_cursor, TO_CURSOR_HIDE, true);
-						load_image(tns.sel);
-					} else {
+					if (sel != tns.sel) {
 						tns_highlight(&tns, tns.sel, false);
 						tns_highlight(&tns, sel, true);
 						tns.sel = sel;
+						firstclick = bev->time;
+						redraw();
+					} else if (bev->time - firstclick <= TO_DOUBLE_CLICK) {
+						mode = MODE_IMAGE;
+						set_timeout(reset_cursor, TO_CURSOR_HIDE, true);
+						load_image(tns.sel);
+						redraw();
+					} else {
+						firstclick = bev->time;
 					}
+				}
+				break;
+			case Button3:
+				if ((sel = tns_translate(&tns, bev->x, bev->y)) >= 0) {
+					files[sel].marked = !files[sel].marked;
+					tns_mark(&tns, sel, files[sel].marked);
 					redraw();
-					break;
 				}
 				break;
 			case Button4:
@@ -493,6 +633,7 @@ void on_buttonpress(XButtonEvent *bev)
 				break;
 		}
 	}
+	prefix = 0;
 }
 
 void run(void)
@@ -515,7 +656,7 @@ void run(void)
 				tns.cnt++;
 			} else {
 				remove_file(tns.cnt, false);
-				if (tns.sel >= tns.cnt)
+				if (tns.sel > 0 && tns.sel >= tns.cnt)
 					tns.sel--;
 			}
 			if (tns.cnt == filecnt)
@@ -563,7 +704,7 @@ void run(void)
 				on_buttonpress(&ev.xbutton);
 				break;
 			case ClientMessage:
-				if ((Atom) ev.xclient.data.l[0] == wm_delete_win)
+				if ((Atom) ev.xclient.data.l[0] == atoms[ATOM_WM_DELETE_WINDOW])
 					return;
 				break;
 			case ConfigureNotify:
@@ -610,7 +751,7 @@ int main(int argc, char **argv)
 	size_t n;
 	ssize_t len;
 	char *filename;
-	const char *homedir;
+	const char *homedir, *dsuffix = "";
 	struct stat fstats;
 	r_dir_t dir;
 
@@ -686,16 +827,25 @@ int main(int argc, char **argv)
 	win_init(&win);
 	img_init(&img, &win);
 
-	if ((homedir = getenv("HOME")) == NULL) {
-		warn("could not locate home directory");
-	} else {
-		len = strlen(homedir) + strlen(INFO_SCRIPT) + 2;
-		info.script = (char*) s_malloc(len);
-		snprintf(info.script, len, "%s/%s", homedir, INFO_SCRIPT);
-		if (access(info.script, X_OK) != 0) {
-			free(info.script);
-			info.script = NULL;
+	if ((homedir = getenv("XDG_CONFIG_HOME")) == NULL || homedir[0] == '\0') {
+		homedir = getenv("HOME");
+		dsuffix = "/.config";
+	}
+	if (homedir != NULL) {
+		char **cmd[] = { &info.cmd, &keyhandler.cmd };
+		const char *name[] = { "image-info", "key-handler" };
+
+		for (i = 0; i < ARRLEN(cmd); i++) {
+			len = strlen(homedir) + strlen(dsuffix) + strlen(name[i]) + 12;
+			*cmd[i] = (char*) s_malloc(len);
+			snprintf(*cmd[i], len, "%s%s/sxiv/exec/%s", homedir, dsuffix, name[i]);
+			if (access(*cmd[i], X_OK) != 0) {
+				free(*cmd[i]);
+				*cmd[i] = NULL;
+			}
 		}
+	} else {
+		warn("could not locate exec directory");
 	}
 	info.fd = -1;
 
