@@ -34,10 +34,10 @@ typedef struct {
 	/* image: */
 	scalemode_t scalemode;
 	float zoom;
-	bool aa;
+	int gamma;
+	int slideshow;
 
 	/* window: */
-	bool fixed_win;
 	bool fullscreen;
 	bool hide_bar;
 	char *geometry;
