@@ -27,7 +27,7 @@
 
 typedef struct {
 	const fileinfo_t *file;
-	Imlib_Image *im;
+	Imlib_Image im;
 	int w;
 	int h;
 	int x;
@@ -47,7 +47,6 @@ typedef struct {
 	int cols;
 	int rows;
 
-	bool alpha;
 	bool dirty;
 } tns_t;
 
@@ -59,6 +58,7 @@ void tns_free(tns_t*);
 bool tns_load(tns_t*, int, const fileinfo_t*, bool, bool);
 
 void tns_render(tns_t*);
+void tns_mark(tns_t*, int, bool);
 void tns_highlight(tns_t*, int, bool);
 
 bool tns_move_selection(tns_t*, direction_t, int);
